from django.http import HttpRequest, HttpResponse

from troncos.profiling.bootstrap.profile import get_profile


def profiling_view(request: HttpRequest) -> HttpResponse:
    profile, headers = get_profile()

    return HttpResponse(profile, headers=headers)



from troncos.profiling import bootstrap, profiler


def start_profiler() -> None:
    if hasattr(bootstrap, "profiler"):
        bootstrap.profiler.stop()

    # Export the profiler so we can introspect it if needed
    bootstrap.profiler = profiler.Profiler()  # type: ignore
    bootstrap.profiler.start()  # type: ignore


start_profiler()



from typing import Any

from troncos.profiling.bootstrap.profile import get_profile


async def profiling_asgi_app(scope: Any, receive: Any, send: Any) -> None:
    profile, headers = get_profile()

    await send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                [header.lower().encode(), value.encode()]
                for header, value in headers.items()
            ],
        }
    )

    await send(
        {
            "type": "http.response.body",
            "body": profile,
        }
    )



import time
from typing import Any, Awaitable, Callable, Iterator, Mapping, MutableMapping

import ddtrace
from ipware import IpWare

try:
    from structlog import get_logger
except ImportError:
    raise RuntimeError(
        "Structlog must be installed to use the asgi logging middleware."
    )


class Headers(Mapping[str, str]):
    """
    An immutable, case-insensitive multidict.
    """

    def __init__(
        self,
        scope: MutableMapping[str, Any],
    ) -> None:
        self._list: list[tuple[bytes, bytes]] = list(scope["headers"])

    def add_client(self, client: tuple[str, int]) -> None:
        """
        The client IP is not stored in the ASGI headers by default.
        Add the client ip to make sure we use it as a fallback if no
        proxy headers are set.
        """
        self._list.append(
            (
                "REMOTE_ADDR".encode("latin-1"),
                f"{client[0]}:{client[1]}".encode("latin-1"),
            )
        )

    def keys(self) -> list[str]:  # type: ignore[override]
        return [key.decode("latin-1") for key, value in self._list]

    def values(self) -> list[str]:  # type: ignore[override]
        return [value.decode("latin-1") for key, value in self._list]

    def items(self) -> list[tuple[str, str]]:  # type: ignore[override]
        return [
            (key.decode("latin-1"), value.decode("latin-1"))
            for key, value in self._list
        ]

    def getlist(self, key: str) -> list[str]:
        get_header_key = key.lower().encode("latin-1")
        return [
            item_value.decode("latin-1")
            for item_key, item_value in self._list
            if item_key == get_header_key
        ]

    def __getitem__(self, key: str) -> str:
        get_header_key = key.lower().encode("latin-1")
        for header_key, header_value in self._list:
            if header_key == get_header_key:
                return header_value.decode("latin-1")
        raise KeyError(key)

    def __contains__(self, key: Any) -> bool:
        get_header_key = key.lower().encode("latin-1")
        for header_key, header_value in self._list:
            if header_key == get_header_key:
                return True
        return False

    def __iter__(self) -> Iterator[Any]:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self._list)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Headers):
            return False
        return sorted(self._list) == sorted(other._list)


class AsgiLoggingMiddleware:
    """
    ASGI application middleware that logs requests.
    """

    def __init__(
        self,
        app: Any,
        logger_name: str | None = None,
    ) -> None:
        self._app = app
        ln = logger_name or "troncos.asgi"
        self._access = get_logger(f"{ln}.access")
        self._error = get_logger(f"{ln}.error")

    async def __call__(
        self,
        scope: dict[str, Any],
        receive: Callable[[Any], Any],
        send: Callable[[dict[str, Any]], Awaitable[Any]],
    ) -> Any:
        if scope["type"] != "http":
            return await self._app(scope, receive, send)

        ipware = IpWare()

        headers = Headers(scope=scope)
        headers.add_client(scope["client"])

        client_ip, _ = ipware.get_client_ip(headers)

        method = scope.get("method")
        path = scope.get("path")
        http_version = scope.get("http_version")
        status = [0]
        start_time = time.perf_counter()

        async def wrapped_send(message: dict[str, Any]) -> None:
            if "status" in message:
                status[0] = message.get("status", 0)
            await send(message)

        log_fn = self._access.info
        extra = {}

        # Earlier implementations of this middleware logged calls in the 'wrapped_send'
        # function above, and made it the responsibility of the trace injection log
        # processor to add the trace/span id to the log entry. The problem is that the
        # ASGI TraceMiddleware defined in ddtrace, sometimes ends the span during the
        # 'send(message)' call, and so the trace injection processor would not see the
        # trace context, and therefore not log the trace/span id.
        #
        # To ensure that the trace information is always logged, we simply inject that
        # information in here, and by doing so we do not have to care about the
        # internals of the ASGI TraceMiddleware in ddtrace.
        if dd_context := ddtrace.tracer.current_trace_context():
            extra["trace_id"] = f"{dd_context.trace_id:x}"
            extra["span_id"] = f"{dd_context.span_id:x}"

        try:
            return await self._app(scope, receive, wrapped_send)
        except Exception as e:
            status[0] = 500
            log_fn = self._error.exception
            raise e
        finally:
            log_fn(
                "ASGI HTTP response",
                http_client_addr=str(client_ip) if client_ip else "NO_IP",
                http_method=method,
                http_path=path,
                http_version=http_version,
                http_status_code=status[0],
                duration=time.perf_counter() - start_time,
                **extra,
            )



from starlette.requests import Request
from starlette.responses import Response

from troncos.profiling.bootstrap.profile import get_profile


async def profiling_view(request: Request) -> Response:
    profile, headers = get_profile()

    return Response(content=profile, headers=headers)



from typing import Any

from ddtrace import constants, ext
from ddtrace.span import Span as DDSpan
from opentelemetry.attributes import BoundedAttributes  # type: ignore
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import (
    _DEFAULT_OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
    _DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    _DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT,
    Event,
    ReadableSpan,
)
from opentelemetry.sdk.util import BoundedList
from opentelemetry.trace import SpanContext, SpanKind, Status, StatusCode
from opentelemetry.trace.span import TraceFlags

_dd_span_ignore_attr = {
    "runtime-id",
    "_sampling_priority_v1",
    "env",
    "version",
    "span.kind",
}

_trace_flags_sampled = TraceFlags(1)


def _span_context(span: DDSpan) -> SpanContext:
    return SpanContext(
        trace_id=span.trace_id,
        span_id=span.span_id,
        is_remote=False,
        trace_flags=_trace_flags_sampled,
    )


def _parent_span_context(dd_span: DDSpan) -> SpanContext | None:
    if dd_span.parent_id:
        if not dd_span._parent:
            # External trace parent
            return SpanContext(dd_span.trace_id, dd_span.parent_id, True)
        else:
            return _span_context(dd_span._parent)

    return None


_span_kind_map = {
    ext.SpanKind.CLIENT: SpanKind.CLIENT,
    ext.SpanKind.SERVER: SpanKind.SERVER,
    ext.SpanKind.PRODUCER: SpanKind.PRODUCER,
    ext.SpanKind.CONSUMER: SpanKind.CONSUMER,
}


def _span_kind(dd_span: DDSpan) -> SpanKind:
    dd_kind = dd_span._meta.get(constants.SPAN_KIND, "none")
    return _span_kind_map.get(dd_kind, SpanKind.INTERNAL)


_dd_span_err_attr_mapping = {
    "error.msg": "exception.message",
    "error.type": "exception.type",
    "error.stack": "exception.stacktrace",
}


def _span_status_and_attributes(
    dd_span: DDSpan, ignore_attrs: set[str]
) -> tuple[Status, list[Event], dict[str, Any]]:
    # Collect all "attributes" from the dd span
    dd_span_attr: dict[str, Any] = {
        **dd_span._meta,  # type: ignore
        **dd_span._metrics,  # type: ignore
        "resource": dd_span.resource,
    }

    otel_attrs = {}
    events: list[Event] = []
    otel_error_attrs = {}

    # Map set OTEL attributes based on DD attributes
    for k, v in dd_span_attr.items():
        if k.startswith("_dd"):
            continue
        otel_err_attr = _dd_span_err_attr_mapping.get(k)
        if otel_err_attr:
            otel_error_attrs[otel_err_attr] = v
        elif k not in ignore_attrs:
            otel_attrs[k] = v

    if otel_error_attrs:
        events.append(
            Event(
                "exception",
                BoundedAttributes(
                    _DEFAULT_OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
                    attributes=otel_error_attrs,
                ),
            )
        )

        status_exp_type = otel_error_attrs.get("exception.type", None)
        status_exp_msg = otel_error_attrs.get("exception.message", None)

        status = Status(
            status_code=StatusCode.ERROR,
            description=f"{status_exp_type}: {status_exp_msg}",
        )
    else:
        status = Status(StatusCode.UNSET)

    return (status, events, otel_attrs)


def _span_resource(dd_span: DDSpan, default_resource: Resource) -> Resource:
    if default_resource.attributes[SERVICE_NAME] == dd_span.service:
        return default_resource

    if not dd_span.service:
        return default_resource

    base_attributes = dict(default_resource.attributes)
    base_attributes[SERVICE_NAME] = dd_span.service

    return Resource(base_attributes)


def default_ignore_attrs() -> set[str]:
    return _dd_span_ignore_attr


def translate_span(
    dd_span: DDSpan, default_resource: Resource, ignore_attrs: set[str]
) -> ReadableSpan:
    """Transelate a ddtrace span to an OTEL span."""
    assert dd_span.duration_ns, "Span not finished."

    status, events, attributes = _span_status_and_attributes(
        dd_span, ignore_attrs=ignore_attrs
    )

    otel_span = ReadableSpan(
        name=dd_span.name,
        context=_span_context(dd_span),
        parent=_parent_span_context(dd_span),
        resource=_span_resource(dd_span, default_resource),
        attributes=BoundedAttributes(
            _DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT, attributes=attributes
        ),
        events=BoundedList.from_seq(_DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT, events),
        kind=_span_kind(dd_span),
        status=status,
        start_time=dd_span.start_ns,
        end_time=dd_span.start_ns + dd_span.duration_ns,
    )

    return otel_span



from typing import Any

import ddtrace

from ._enums import Exporter
from ._writer import OTELWriter

__all__ = ["Exporter"]


def configure_tracer(
    *,
    enabled: bool,
    endpoint: str,
    service_name: str,
    service_attributes: dict[str, Any] | None = None,
    exporter: Exporter = Exporter.HTTP,
) -> None:
    """Configure ddtrace to write traces to the otel tracing backend."""

    # Initialize our custom writer used to handle ddtrace spans.
    writer = OTELWriter(
        service_name=service_name,
        service_attributes=service_attributes,
        endpoint=endpoint,
        exporter=exporter,
    )

    ddtrace.tracer.configure(writer=writer, enabled=enabled)



from typing import Any

from ddtrace.internal.writer.writer import TraceWriter
from ddtrace.span import Span
from opentelemetry.sdk.resources import Resource

from ._enums import Exporter
from ._otel import get_otel_span_processors
from ._span import default_ignore_attrs, translate_span


class OTELWriter(TraceWriter):
    def __init__(
        self,
        service_name: str,
        service_attributes: dict[str, Any] | None,
        endpoint: str,
        exporter: Exporter,
    ) -> None:
        self.service_name = service_name
        self.service_attributes = service_attributes
        self.endpoint = endpoint
        self.exporter = exporter

        self.otel_span_processors = get_otel_span_processors(
            endpoint=endpoint, exporter=exporter
        )
        self.otel_default_resource = Resource.create(
            {"service.name": service_name, **(service_attributes or {})}
        )
        self.otel_ignore_attrs = (
            set(self.otel_default_resource.attributes.keys()) | default_ignore_attrs()
        )

    def recreate(self) -> "OTELWriter":
        return self.__class__(
            self.service_name,
            self.service_attributes,
            self.endpoint,
            exporter=self.exporter,
        )

    def write(self, spans: list[Span] | None = None) -> None:
        if not spans:
            return

        filtered_spans = [
            span
            for span in spans
            # ddtrace use span.sampled == False to drop spans.
            if span.sampled
            # ddtrace uses sampling_priority > 0 to indicate that we
            # want to ingest the span.
            and (
                span.context.sampling_priority is None
                or span.context.sampling_priority > 0
            )
        ]

        if not filtered_spans:
            return

        transelated_spans = [
            translate_span(
                span,
                default_resource=self.otel_default_resource,
                ignore_attrs=self.otel_ignore_attrs,
            )
            for span in filtered_spans
        ]

        for span_processor in self.otel_span_processors:
            for span in transelated_spans:
                span_processor.on_end(span)

    def stop(self, timeout: float | None = None) -> None:
        for span_processor in self.otel_span_processors:
            span_processor.force_flush(
                timeout_millis=int(timeout * 1000) if timeout else 30000
            )
            span_processor.shutdown()

    def flush_queue(self) -> None:
        for span_processor in self.otel_span_processors:
            span_processor.force_flush()



import os
import sys

from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPSpanExporter,
)
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SimpleSpanProcessor,
    SpanExporter,
)
from structlog import get_logger

from ._enums import Exporter

try:
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
        OTLPSpanExporter as GRPCSpanExporter,
    )
except ImportError:
    GRPCSpanExporter = None  # type: ignore


logger = get_logger()


def _bool_from_string(s: str) -> bool:
    return s.lower() in ["1", "true", "yes"]


def get_otel_span_processors(
    *, endpoint: str, exporter: Exporter
) -> list[SpanProcessor]:
    """
    Build a list of span processors to use to process otel spans.
    """

    span_processors: list[SpanProcessor] = []
    span_exporter: SpanExporter

    # Exporter
    if exporter == Exporter.HTTP:
        span_exporter = HTTPSpanExporter(endpoint=endpoint)
    elif exporter == Exporter.GRPC:
        if GRPCSpanExporter is None:
            raise RuntimeError(
                "opentelemetry-exporter-otlp-proto-grpc needs to be installed "
                "to use the GRPC exporter."
            )

        span_exporter = GRPCSpanExporter(endpoint=endpoint)
    else:
        raise RuntimeError("Unsupported span exporter.")

    span_processors.append(BatchSpanProcessor(span_exporter))

    # Setup OTEL debug processor
    otel_trace_debug = _bool_from_string(os.environ.get("OTEL_TRACE_DEBUG", "false"))
    otel_trace_debug_file = os.environ.get("OTEL_TRACE_DEBUG_FILE")
    if otel_trace_debug:
        logger.info(f"OTEL debug processor to {otel_trace_debug_file or 'stdout'}")
        debug_out = (
            sys.stdout
            if not otel_trace_debug_file
            else open(otel_trace_debug_file, "w")
        )
        span_processors.append(SimpleSpanProcessor(ConsoleSpanExporter(out=debug_out)))

    return span_processors



from troncos.profiling import bootstrap


def get_profile() -> tuple[bytes, dict[str, str]]:
    """Get the latest profile."""

    headers = {
        "Content-Type": "application/octet-stream",
        "Content-Disposition": 'attachment; filename="python.pprof"',
        "X-Content-Type-Options": "nosniff",
    }

    if hasattr(bootstrap, "profiler"):
        return bootstrap.profiler.get_profile(), headers

    return b"", headers



import time
from typing import Any

from celery import signals

try:
    from structlog import get_logger
except ImportError:
    raise RuntimeError("Structlog must be installed to use the celery logging signals.")

logger = get_logger("troncos.celery.task")


def connect_troncos_logging_celery_signals() -> None:
    """
    Log a message every time a task is complete.
    """

    signals.task_prerun.connect(_prerun, weak=True)
    signals.task_postrun.connect(_postrun, weak=True)


def _prerun(sender: Any, task_id: Any, task: Any, *args: Any, **kwargs: Any) -> None:
    task.__troncos_start_time = time.perf_counter()


def _postrun(sender: Any, task_id: Any, task: Any, *args: Any, **kwargs: Any) -> None:
    started_time = getattr(task, "__troncos_start_time", None)

    extra = {}

    if started_time:
        extra["duration"] = time.perf_counter() - started_time

    logger.info("Celery task post-run", task=task.name, state=kwargs["state"], **extra)



from typing import Any

from ddtrace.profiling import exporter
from ddtrace.profiling.exporter import pprof  # type: ignore
from ddtrace.profiling.profiler import (
    Profiler as DDProfiler,
    _ProfilerInstance as DDProfilerInstance,
)


class _PprofExporter(pprof.PprofExporter):  # type: ignore[misc]
    pprof = ""

    @staticmethod
    def _map_frames(events):  # type: ignore[no-untyped-def]
        """
        This function takes the profiling events, and iterates over all the frames,
        formatting the data for phlare.
        """

        for _, e1 in events.items():
            for e2 in e1:
                e2.frames = list(e2.frames)
                for i in range(len(e2.frames)):
                    frame = e2.frames[i]
                    file = frame[0].split("/lib/", 1)[-1]
                    loc = f"{file}:{frame[1]}:{frame[2]}"
                    e2.frames[i] = (frame[0], frame[1], loc, frame[3])

    def export(self, events, start_time_ns, end_time_ns):  # type: ignore[no-untyped-def] # noqa: E501
        self._map_frames(events)  # type: ignore[no-untyped-call]
        pprof_profile, _ = super(_PprofExporter, self).export(
            events, start_time_ns, end_time_ns
        )

        self.pprof = pprof_profile.SerializeToString()


class _ProfilerInstance(DDProfilerInstance):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.exporter = _PprofExporter()

        super().__init__(*args, **kwargs)

    def _build_default_exporters(self) -> list[exporter.Exporter]:
        return [self.exporter]


class Profiler(DDProfiler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._profiler = _ProfilerInstance(*args, **kwargs)

    def get_profile(self) -> str:
        return self._profiler.exporter.pprof  # type: ignore



import asyncio
import inspect
import logging
from collections.abc import Generator
from contextlib import contextmanager
from functools import wraps
from types import FunctionType
from typing import Awaitable, Callable, ParamSpec, Type, TypeVar, cast, overload

import ddtrace
from ddtrace.span import Span

_TRACE_IGNORE_ATTR = "_trace_ignore"

TClass = TypeVar("TClass")

P = ParamSpec("P")
R = TypeVar("R")


@contextmanager
def trace_block(
    name: str,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Generator[Span, None, None]:
    """
    Trace a code block using a with statement. Example:

    with trace_block("cool.block", resource="data!", attributes={"some": "attribute"}):
        time.sleep(1)
    """

    with ddtrace.tracer.trace(
        name=name,
        resource=resource,
        service=service,
        span_type=span_type,
    ) as span:
        span.set_tags(attributes)
        yield span


def _trace_function(
    f: Callable[P, R],
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R]:
    if hasattr(f, _TRACE_IGNORE_ATTR):
        return f

    # Async function
    if inspect.iscoroutinefunction(f):

        @wraps(f)
        async def traced_func_async(*args: P.args, **kwargs: P.kwargs) -> R:
            with trace_block(
                name=name or f"{f.__module__}.{f.__qualname__}",
                resource=resource,
                service=service,
                span_type=span_type,
                attributes=attributes,
            ):
                awaitable_func = cast(Callable[P, Awaitable[R]], f)
                return await awaitable_func(*args, **kwargs)

        return cast(Callable[P, R], traced_func_async)

    else:
        # "Regular" function
        @wraps(f)
        def traced_func(*args: P.args, **kwargs: P.kwargs) -> R:
            with trace_block(
                name=name or f"{f.__module__}.{f.__qualname__}",
                resource=resource,
                service=service,
                span_type=span_type,
                attributes=attributes,
            ):
                return f(*args, **kwargs)

        return traced_func


@overload
def trace_function(
    fn: Callable[P, R],
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R]:
    ...


@overload
def trace_function(
    fn: None = None,
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    ...


def trace_function(
    fn: Callable[P, R] | None = None,
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R] | Callable[[Callable[P, R]], Callable[P, R]]:
    """
    This decorator adds tracing to a function. Example:

    @trace_function
    def myfunc1()
        return "This will be traced"

    @trace_function(service="custom_service")
    def myfunc2()
        return "This will be traced as a custom service"
    """

    if fn and (callable(fn) or asyncio.iscoroutinefunction(fn)):
        return _trace_function(fn, name, resource, service, span_type, attributes)
    else:
        # No args
        def _inner(f: Callable[P, R]) -> Callable[P, R]:
            return _trace_function(f, name, resource, service, span_type, attributes)

        return _inner


@overload
def trace_class(
    klass: None = None,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[[Type[TClass]], Type[TClass]]:
    ...


@overload
def trace_class(
    klass: Type[TClass],
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Type[TClass]:
    ...


def trace_class(
    klass: Type[TClass] | None = None,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Type[TClass] | Callable[[Type[TClass]], Type[TClass]]:
    """
    This decorator adds a tracing decorator to every method of the decorated class. If
    you don't want some methods to be traced, you can add the 'trace_ignore' decorator
    to them. Example:

    @trace_class
    class MyClass1:

        def m1(self):
            return "This will be traced"

        @trace_ignore
        def m2(self):
            return "This will not be traced"


    @trace_class(service="custom_service")
    class MyClass2:

        def m3(self):
            return "This will be traced as a custom service"
    """

    def _class_decorator(cls: Type[TClass]) -> Type[TClass]:
        for key, value in cls.__dict__.items():
            if key.startswith("_"):
                continue
            if not (
                isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)
            ):
                continue

            logging.getLogger(__name__).debug(f"Tracing function {cls.__name__}.{key}")
            setattr(
                cls,
                key,
                _trace_function(
                    value,
                    name=None,
                    resource=resource,
                    service=service,
                    span_type=span_type,
                    attributes=attributes,
                ),
            )
        return cls

    if klass:
        return _class_decorator(klass)
    else:
        return _class_decorator


def trace_module(
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> None:
    """
    This function adds a tracing decorator to every function of the calling module. If
    you don't want some functions to be traced, you can add the 'trace_ignore' decorator
    to them. Example:

    # Start of module

    def my_function():
        return "This func will be traced"

    @trace_ignore
    def my_function():
        return "This func will not be traced"

    trace_module()

    # End of module
    """

    frame = inspect.stack()[1].frame
    scope = frame.f_locals
    module_name = scope.get("__name__", "unknown")

    for key, value in list(frame.f_locals.items()):
        if key.startswith("_"):
            continue
        if not (isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)):
            continue
        if getattr(value, "__module__", None) != module_name:
            continue

        logging.getLogger(__name__).debug(f"Tracing function {module_name}.{key}")
        scope[key] = _trace_function(
            value,
            name=None,
            resource=resource,
            service=service,
            span_type=span_type,
            attributes=attributes,
        )


def trace_ignore(f: Callable[P, R]) -> Callable[P, R]:
    """
    Decorator to disable automatic tracing of functions.
    See 'trace_module' and 'trace_class'.
    """

    setattr(f, _TRACE_IGNORE_ATTR, ())
    return f



from importlib import metadata

# Get version using this method
# https://github.com/python-poetry/poetry/issues/273#issuecomment-1103812336

OTEL_LIBRARY_NAME = "troncos"
OTEL_LIBRARY_VERSION = metadata.version(__package__)

del metadata  # avoids polluting the results of dir(__package__)



import time
from typing import Any

from asgiref.sync import iscoroutinefunction
from django.http import HttpRequest, HttpResponse
from django.utils.decorators import sync_and_async_middleware
from ipware import IpWare

try:
    from structlog import get_logger
except ImportError:
    raise RuntimeError(
        "Structlog must be installed to use the asgi logging middleware."
    )


@sync_and_async_middleware
def DjangoLoggingMiddleware(get_response):  # type: ignore
    """
    Django middleware that logs requests.
    """

    access = get_logger("troncos.django.access")
    error = get_logger("troncos.django.error")

    ipware = IpWare()

    def extract_request_data(request: HttpRequest) -> tuple[dict[str, Any], float]:
        start_time = time.perf_counter()

        client_ip, _ = ipware.get_client_ip(request.META)

        request_data = {
            "http_method": request.method,
            "http_path": request.path,
            "http_client_addr": str(client_ip) if client_ip else "NO_IP",
        }

        return request_data, start_time

    def log_response(
        *, response: HttpResponse, request_data: dict[str, Any], start_time: float
    ) -> None:
        http_status_code = response.status_code

        logger_method = access.info if http_status_code < 500 else error.error

        logger_method(
            "Django HTTP response",
            http_status_code=http_status_code,
            duration=time.perf_counter() - start_time,
            **request_data,
        )

    if iscoroutinefunction(get_response):

        async def middleware(request: HttpRequest) -> HttpResponse:
            request_data, start_time = extract_request_data(request)

            response = await get_response(request)

            log_response(
                response=response,
                request_data=request_data,
                start_time=start_time,
            )

            return response

    else:

        def middleware(request: HttpRequest) -> HttpResponse:  # type: ignore
            request_data, start_time = extract_request_data(request)

            response = get_response(request)

            log_response(
                response=response,
                request_data=request_data,
                start_time=start_time,
            )

            return response

    return middleware



from ddtrace import tracer
from structlog.processors import LogfmtRenderer as LogFmt
from structlog.types import EventDict, WrappedLogger


def trace_injection_processor(
    _logger: WrappedLogger, _log_method: str, event_dict: EventDict
) -> EventDict:
    """
    Simple logging processor that adds a trace_id to the log record if available.
    """

    # Try to get context from tracer
    dd_context = tracer.current_trace_context()

    # Add context to log record if exists
    if dd_context:
        event_dict["trace_id"] = f"{dd_context.trace_id:x}"
        event_dict["span_id"] = f"{dd_context.span_id:x}"

    return event_dict


class LogfmtRenderer(LogFmt):
    """
    A structlog Logfmt renderer that does not produce new lines
    """

    def __call__(self, _: WrappedLogger, __: str, event_dict: EventDict) -> str:
        return super().__call__(_, __, event_dict).replace("\n", " ")



import logging
import logging.config

import structlog

from troncos.contrib.structlog.processors import (
    LogfmtRenderer,
    trace_injection_processor,
)

try:
    from structlog_sentry import SentryProcessor
except ImportError:
    SentryProcessor = None

shared_processors: list[structlog.types.Processor] = [
    # Add the name of the logger to event dict.
    structlog.stdlib.add_logger_name,
    # Add log level to event dict.
    structlog.stdlib.add_log_level,
    # Add a timestamp in ISO 8601 format.
    structlog.processors.TimeStamper(fmt="iso"),
    # If the "exc_info" key in the event dict is either true or a
    # sys.exc_info() tuple, remove "exc_info" and render the exception
    # with traceback into the "exception" key.
    structlog.processors.format_exc_info,
    trace_injection_processor,
]

if SentryProcessor is not None:
    shared_processors.append(
        SentryProcessor(level=logging.INFO, event_level=logging.ERROR)
    )


def configure_structlog(
    *, configure_logging: bool = True, format: str = "text", level: str = "INFO"
) -> None:
    """
    Helper method to configure Structlog.

    Using this is not required, you can configure Structlog
    manually in your application.

    configure_logging=True lets you use structlog to render logs from
    logging.getLogger, this is used to get an unified log output.
    """

    processor: structlog.types.Processor

    if format == "text":
        processor = structlog.dev.ConsoleRenderer(colors=True)
    elif format == "json":
        processor = structlog.processors.JSONRenderer()
    elif format == "logfmt":
        processor = LogfmtRenderer()
    else:
        raise RuntimeError(f"Invalid log format {format}")

    if configure_logging:
        config = {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": processor,
                    "foreign_pre_chain": shared_processors,
                },
            },
            "handlers": {
                "default": {
                    "formatter": "default",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stderr",
                },
            },
            "loggers": {
                "": {
                    "handlers": ["default"],
                    "level": level,
                    "propagate": True,
                },
            },
        }

        logging.config.dictConfig(config)

    structlog_processors: list[structlog.types.Processor] = [
        # Merge contextvars into the event dict.
        structlog.contextvars.merge_contextvars,
        # If log level is too low, abort pipeline and throw away log entry.
        structlog.stdlib.filter_by_level,
        # Add shared processors to the processor chain.
        *shared_processors,
        # Perform %-style formatting.
        structlog.stdlib.PositionalArgumentsFormatter(),
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If some value is in bytes, decode it to a unicode str.
        structlog.processors.UnicodeDecoder(),
        # Add callsite parameters.
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.LINENO,
            }
        ),
    ]

    if configure_logging:
        # Prepare event dict for `ProcessorFormatter`.
        structlog_processors.append(
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter
        )
    else:
        structlog_processors.append(processor)

    structlog.configure(
        processors=structlog_processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )



import pyroscope
from ddtrace import config, tracer
from ddtrace.internal.hostname import get_hostname


def start_py_spy_profiler(
    *,
    server_address: str,
    tags: dict[str, str] | None = None,
    enable_logging: bool = False,
    auth_token: str = "",
    basic_auth_username: str = "",
    basic_auth_password: str = "",
) -> None:
    """Start the py-spy continuous profiler."""

    # Use the default ddtrace service name.
    app_name = config.service or "unknown-service"

    profiler_tags = {
        "app": app_name,
        "env": config.env,
        "version": config.version,
        "instance": get_hostname(),
        # Use tags from the global tracer.
        **tracer._tags,
    }

    if tags:
        profiler_tags.update(tags)

    pyroscope.configure(
        app_name=app_name,
        tags=profiler_tags,
        server_address=server_address,
        sample_rate=100,
        detect_subprocesses=True,
        oncpu=False,
        gil_only=False,
        enable_logging=enable_logging,
        auth_token=auth_token,
        basic_auth_username=basic_auth_username,
        basic_auth_password=basic_auth_password,
    )
